/*
   glens
   Version 0.1.
   Gravitational lens simulation
   Copyright (C) PG Sprimont 2002

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <GL/glut.h>
#include <glui.h>

#include "gl.h"
#include "phys.h"
#include "ui.h"
#include "mat.h"
#include "../config.h"


//#define DEBUG
/*-------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------*/

extern int ndefl;

extern int nrays;                              	
extern defl* dlist;
extern light *ray, *Ray;

extern int moving;
extern float capture_radius;
extern int accurate;

extern float* angles;
extern int nangles;

/*-------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------*/

int catched=0; //nombre de photons intercepts

opt rt_opt={50.0,15.0,1.0,1,1,1,1,1,1,1,0};

//global view

view glob_view={
                {1,0,0,0,   //Matrice de rotation
                 0,1,0,0,
                 0,0,1,0,
                 0,0,0,1}

                ,{0,0,0}};   //Translations

//observer view

view obs_view={
                {1,0,0,0,   //Matrice de rotation
                 0,1,0,0,
                 0,0,1,0,
                 0,0,0,1}

                ,{0,0,0}};   //Translations


GLfloat MassFactor=1.0;   //Grossissement des reprsentations des dflecteurs

static GLint T0 = 0;
static GLint Frames = 0;
int nseg=500; //nombre de pas maximum/rayon

GLfloat vuefrac=1.0;
GLfloat globw, globh, globx, globy;
GLfloat obsplanelarg=6.0;  //dimensions du carr (oeil)
int radar1, radar2;
float pixsize=2.0;

GLfloat radius=3.0;			//rayon de l'orbite du dflecteur 2
static GLfloat angle = 0.0;

int win_main; //fenetre openGL principale

/*-------------------------------------------------------------------------------*/
void drawaxes( float scale )
/*-------------------------------------------------------------------------------*/
{

  glPushMatrix();

  glTranslatef(-0.05*scale,-0.05*scale,-0.05*scale);
  glScalef( scale, scale, scale );

  glBegin( GL_LINES );

  glColor3f( 1.0, 0.0, 0.0 );
  glVertex3f( 0.0, 0.0, 0.0 );  glVertex3f( 1.0, 0.0, 0.0 ); /* X axis */
  glColor3f( 0.0, 1.0, 0.0 );
  glVertex3f( 0.0, 0.0, 0.0 );  glVertex3f( 0.0, 1.0, 0.0 ); /* Y axis */
  glColor3f( 0.0, 1.0, 1.0 );
  glVertex3f( 0.0, 0.0, 0.0 );  glVertex3f( 0.0, 0.0, 1.0 ); /* Z axis  */
  glEnd();

  glColor3f( 1.0, 1.0, 1.0 );

  glRasterPos3f(0.0,0.0,1.0);
  glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, 90);
  glRasterPos3f(0.0,1.0,0.0);
  glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, 89);
  glRasterPos3f(1.0,0.0,0.0);
  glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, 88);
  glRasterPos3f(0.05,0.05,0.05);

  glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, 48);
  glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, 44);
  glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, 48);
  glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, 44);
  glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, 48);


  glPopMatrix();
}
/*-------------------------------------------------------------------------------*/
void drawrays() //Raytracing
/*-------------------------------------------------------------------------------*/
{

  restoreraysall();
  register int k;

	for(k=0;k<nrays;k++)
  {
	
   glColor3f(0.0,0.4,0.8);
	int count=0;
	
    if(rt_opt.disprays)
    {

		  glBegin(GL_LINE_STRIP);
	
		  while( ((! rt_opt.boundary_lateral) ||    //Essaye de comprendre !! =))
                      ((ray[k].P[1]>=0.0)&&(ray[k].P[1]<=rt_opt.Boxl)
             &&   (ray[k].P[0]>=0.0)&&(ray[k].P[0]<=rt_opt.Boxl)))
             &&   ((! rt_opt.boundary_front) ||
                      ((ray[k].P[2]<=0.0)&&(ray[k].P[2]>=-rt_opt.BoxL)))
		    		   && (count<nseg) )
		  {
			  ++count;

//        if((ray[k].flagged&&moving) || !moving)
				  glVertex3f(ray[k].P[0],ray[k].P[1],ray[k].P[2]);
			  ray[k].dev(); //fonction de dviation physique
        if(!ray[k].flagged)
        {
          if(ray[k].intersect())
          {   //test d'interception plan observateur
		        ray[k].flagged=true;  //et mmorisation de l'angle si succes
            glColor3f(1.0,0.0,0.0);
          }
        }
      }

		  glEnd();
    }
    else
    {
      while( ((! rt_opt.boundary_lateral) ||
                      ((ray[k].P[1]>=0.0)&&(ray[k].P[1]<=rt_opt.Boxl)
             &&   (ray[k].P[0]>=0.0)&&(ray[k].P[0]<=rt_opt.Boxl)))
             &&   ((! rt_opt.boundary_front) ||
                      ((ray[k].P[2]<=0.0)&&(ray[k].P[2]>=-rt_opt.BoxL)))
		    		   && (count<nseg) )
		  {

			  ++count;
			  ray[k].dev();
        if(!ray[k].flagged)
        {
          ray[k].intersect();
          ray[k].flagged=true;
        }
		  }

    }
  }
}
/*-------------------------------------------------------------------------------*/
void drawbox() //Dessin de la boite (matrialisant la zone de calcul)
/*-------------------------------------------------------------------------------*/
{
	glColor3f(.6,.2,.2);

	glBegin(GL_LINE_LOOP);	
	glVertex3f(0.0,0.0,0.0);
	glVertex3f(0.0,rt_opt.Boxl,0.0);
	glVertex3f(rt_opt.Boxl,rt_opt.Boxl,0.0);
  glVertex3f(rt_opt.Boxl,0.0,0.0);
	glEnd();

	glBegin(GL_LINE_LOOP);	
	glVertex3f(0.0,0.0,-rt_opt.BoxL);
	glVertex3f(0.0,rt_opt.Boxl,-rt_opt.BoxL);
	glVertex3f(rt_opt.Boxl,rt_opt.Boxl,-rt_opt.BoxL);
  glVertex3f(rt_opt.Boxl,0.0,-rt_opt.BoxL);
	glEnd();

	glBegin(GL_LINES);
	glVertex3f(0.0,0.0,0.0);
	glVertex3f(0.0,0.0,-rt_opt.BoxL);
	
	glVertex3f(0.0,rt_opt.Boxl,0.0);
	glVertex3f(0.0,rt_opt.Boxl,-rt_opt.BoxL);

	glVertex3f(rt_opt.Boxl,rt_opt.Boxl,0.0);
	glVertex3f(rt_opt.Boxl,rt_opt.Boxl,-rt_opt.BoxL);

	glVertex3f(rt_opt.Boxl,0.0,0.0);
	glVertex3f(rt_opt.Boxl,0.0,-rt_opt.BoxL);
	glEnd();

}
/*-------------------------------------------------------------------------------*/
void drawdefl()             //Dessin des dflecteurs ( amliorer)
/*-------------------------------------------------------------------------------*/
{
 	int i;

	GLfloat x, y;
	
	glColor3f(0.0,0.8,0.2);

	for(i=0;i<ndefl;i++){
		if(dlist[i].active)
    {
      if(dlist[i].orbit)
      {
        x= dlist[i].P[0]+radius*cos(angle);
	      y= dlist[i].P[1]+radius*sin(angle);
      }
      glPushMatrix();
		  glTranslatef(dlist[i].P[0],dlist[i].P[1],dlist[i].P[2]);
		  glutSolidSphere(MassFactor*dlist[i].m, 10,10);
		  glPopMatrix();
    }
	}
}
/*-------------------------------------------------------------------------------*/
static void drawradar(float r, int steps, int rads)
/*-------------------------------------------------------------------------------*/
{
   int i,j;
    float dr=r/rads;

    char text[]="90";
    char text2[]="-90";

    glColor3f(0.8,0.3,0.3);

    glRasterPos2f(-r-dr,0.0);
    for(i=0;i<(int)strlen(text2);i++)
      glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12,text2[i]);
    glRasterPos2f(0.0,r+dr);
    for(i=0;i<(int)strlen(text);i++)
      glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12,text[i]);
    glRasterPos2f(r+dr,0.0);
    for(i=0;i<(int)strlen(text);i++)
      glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12,text[i]);
    glRasterPos2f(0.0,-r-dr);
    for(i=0;i<(int)strlen(text2);i++)
      glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12,text2[i]);


    glColor3f(0.8,0.8,0.8);
    glEnable(GL_LINE_STIPPLE);
    glBegin(GL_LINES);
    glVertex3f(-r-dr,0.0,0.0);
    glVertex3f(r+dr,0.0,0.0);
    glVertex3f(0.0,r+dr,0.0);
    glVertex3f(0.0,-r-dr,0.0);
    glEnd();

    glDisable(GL_LINE_STIPPLE);

   glColor3f(0.4,0.4,0.2);

   for(j=0;j<rads;j++)
   {
     for (i=0;i<steps;i++){
      glBegin(GL_LINE_STRIP);
      glVertex2f(j*dr*cos((2*M_PI*i)/steps), j*dr*sin((2*M_PI*i)/steps));
      glVertex2f(j*dr*cos((2*M_PI*(i+1))/steps),j*dr*sin((2*M_PI*(i+1))/steps));
      glEnd();
    }
   }

    glBegin(GL_LINES);
    r-=dr;
    for(j=0;j<=rads;j++)
    {

      glVertex2f(-r*cos(j*M_PI/rads),-r*sin(j*M_PI/rads));
      glVertex2f(r*cos(j*M_PI/rads),r*sin(j*M_PI/rads));
    }
    glEnd();

}
/*-------------------------------------------------------------------------------*/
static void drawobsview(float l)
/*-------------------------------------------------------------------------------*/
{
  static float changed=0;

  float larg=0.8;

  if((changed-l)!=0)
  {
    changed=l;
    radar2=glGenLists(1);
    glNewList(radar2, GL_COMPILE);
    glColor3f(0.4,0.4,0.4);
    drawradar(larg/2.0,12,12);
    glEndList();
  }
	
 	glColor3f(1.0,1.0,1.0);
  glBegin(GL_LINES);
  glVertex3f(0.0,0.0,0.0);
  glVertex3f(l,0.0,0.0);
  glVertex3f(0.0,0.01,0.0);
  glVertex3f(l,0.01,0.0);
  glEnd();

	glColor3f(1.0,1.0,0.0);
  glPushMatrix();
  glTranslatef(l/2,0.5,0.0);
  glCallList(radar2);
	
  glColor3f(0.0,0.6,0.9);
  float fact=larg/2;

  unsigned int i;
  #ifdef DEBUG
  printf("\nnangles=%d",nangles);
  fflush(stdout);
  #endif

  if(nangles>0)
  {
    #ifdef DEBUG
    printf("\nTracage...");
    #endif
 	  glBegin(GL_POINTS);
	  for(i=0;i< (unsigned int) nangles;i++)    //projection "radar"
	  	glVertex3f(angles[2*i]*fact, angles[2*i+1]*fact, 0.2f);
      //glVertex3f((GLfloat)(angles[2*i]*fact*cos(angles[2*i+1])),
      //           (GLfloat)(angles[2*i]*fact*sin(angles[2*i+1])),0.0f);
	  glEnd();
  }

  //printf("\n%d",nangles);
  glPopMatrix();
}
/*-------------------------------------------------------------------------------*/
static void drawobsplane()
/*-------------------------------------------------------------------------------*/
{

  float larg = 2*capture_radius;

  glColor3f(0.1,0.6,0.2);

  if(accurate)
    drawradar(larg/2,12,12);
  else
    glutWireSphere(capture_radius, 20, 20);

  glColor3f(1.0,1.0,0.0);
  glBegin(GL_LINE_LOOP);
  glVertex3f(-larg/2,-larg/2,0.0);
	glVertex3f(-larg/2,larg/2,0.0);
	glVertex3f(larg/2,larg/2,0.0);
  glVertex3f(larg/2,-larg/2,0.0);
  glEnd();

  drawaxes(1.0);

}
/*-------------------------------------------------------------------------------*/
static void draw(void)    //Affichage gnral d'un cran
/*-------------------------------------------------------------------------------*/
{

  static char text[15];//="--------------";
  int i;

  //Rapport gomtrique de la sous-fenetre GL
  GLfloat hr = (GLfloat) globw / (GLfloat) globh / vuefrac;
  //Passage viewport Bas (raytracing) dfinie par vuefrac
  glViewport((GLint) globx, (GLint) globy, (GLint) globw, (GLint) (globh*vuefrac));
  //Projection perspective
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-hr, hr, -1.0, 1.0, 1.0, 200.0);

  glMatrixMode(GL_MODELVIEW);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glLoadIdentity();
  glPushMatrix();

  view temp;

  if(moving)
  {
    drawobsplane();
    inverse(&obs_view, &temp);
    glMultMatrixf(temp.rot);
    glTranslatef(-obs_view.z[0],-obs_view.z[1],-obs_view.z[2]);
  }
  else
  {
   glTranslatef(-glob_view.z[0],-glob_view.z[1],-glob_view.z[2]);

   glTranslatef(rt_opt.Boxl/2,rt_opt.Boxl/2,-rt_opt.BoxL/2);
   glMultMatrixf(glob_view.rot);
   glTranslatef(-rt_opt.Boxl/2,-rt_opt.Boxl/2,rt_opt.BoxL/2);

   glPushMatrix();
   glTranslatef(obs_view.z[0],obs_view.z[1],obs_view.z[2]);
   glMultMatrixf(obs_view.rot);
   drawobsplane();
   glPopMatrix();

  }

  if(rt_opt.dispbox)
    drawbox();
  if(rt_opt.dispaxes)
    drawaxes(1.0);
  if(rt_opt.dispdefl)
  	drawdefl();

  drawrays();

  glPopMatrix();


  if(rt_opt.dispobsview && (vuefrac!=1))
  {
  //Passage viewport Haut (Observateur)
    hr = (GLfloat) globw / (GLfloat) globh / (1.0-vuefrac);
    glViewport((GLint) globx, (GLint) (globy+globh*vuefrac), (GLint) globw, (GLint) (globh*(1.0-vuefrac)));
  //Projection perspective
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0.0,hr,0.0,1.0);
  //Dessin vue suprieure
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glPushMatrix();
    glRasterPos3f(0.1,0.1,0.1);
    glColor3f(0.7,0.7,0.7);
    if(rt_opt.dispstats)
    for(i=0;i<(int)strlen(text);i++)
      glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12,text[i]);

    drawobsview(hr);
    glPopMatrix();
  }

  //fin du dessin, swap de la mmoire video
  glutSwapBuffers();
  //Stats : fps
  if(rt_opt.dispstats)
  {
    Frames++;
    GLint t = glutGet(GLUT_ELAPSED_TIME);
    if (t - T0 >= 1000)
    {
      GLint fps = (int) ( Frames*1.0 / (t - T0) * 1000.0) ;
      sprintf(text,"%d FPS\n", fps);
      T0 = t;
      Frames = 0;
    }
  }
}
/*-------------------------------------------------------------------------------*/
static void idle(void)
/*-------------------------------------------------------------------------------*/
{
  glutSetWindow(win_main);
  angle += 0.1;             //Angle de rotation du satellite (dlist[1])
  glutPostRedisplay();
}
/*-------------------------------------------------------------------------------*/
static void key(unsigned char k, int x, int y) //controle clavier
/*-------------------------------------------------------------------------------*/
{
  switch (k) {
  case 'z':
    break;
  default:
    return;
  }
  glutPostRedisplay();
}

/*-------------------------------------------------------------------------------*/
static void special(int k, int x, int y)
/*-------------------------------------------------------------------------------*/
{
  switch (k) {
  case GLUT_KEY_UP:
    break;
  default:
    return;
  }
  glutPostRedisplay();
}
/*-------------------------------------------------------------------------------*/
static void reshape(int width, int height)       //resizing de la fenetre
/*-------------------------------------------------------------------------------*/
{
  int x,y,w,h;
  GLUI_Master.get_viewport_area(&x,&y,&w,&h);
  globw=w;
  globh=h;
  globx=x;
  globy=y;
}
/*-------------------------------------------------------------------------------*/
void init(int argc, char *argv[])       //initialisation graphique
/*-------------------------------------------------------------------------------*/
{
  glEnable(GL_DEPTH_TEST);
  //glEnable(GL_NORMALIZE);

  glLineStipple(1, 0xAAAA);

  glPointSize(pixsize);

  if (argc > 1 && strcmp(argv[1], "-info")==0) {
     printf("GL_RENDERER   = %s\n", (char *) glGetString(GL_RENDERER));
     printf("GL_VERSION    = %s\n", (char *) glGetString(GL_VERSION));
     printf("GL_VENDOR     = %s\n", (char *) glGetString(GL_VENDOR));
     printf("GL_EXTENSIONS = %s\n", (char *) glGetString(GL_EXTENSIONS));
  }

  radar1=glGenLists(1);
  glNewList(radar1, GL_COMPILE);
  glColor3f(0.2,0.2,0.2);
  drawradar(obsplanelarg/2.0,12,12);
  glEndList();
}
/*-------------------------------------------------------------------------------*/
int glmain(int argc, char *argv[])
/*-------------------------------------------------------------------------------*/
{

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);

  glutInitWindowPosition(100, 100);
  glutInitWindowSize(650, 800);
  char title[256];
  sprintf(title, "%s -- Version %s", PACKAGE, VERSION);

  win_main = glutCreateWindow(title);

  GLUI_Master.set_glutDisplayFunc(draw);
  GLUI_Master.set_glutReshapeFunc(reshape);
  GLUI_Master.set_glutKeyboardFunc(key);
  GLUI_Master.set_glutSpecialFunc(special);

  GLUI_Master.set_glutIdleFunc(idle);


  init(argc, argv);

  init_glui();

  glutMainLoop();

  return 0;
}
