/*
   glens
   Version 0.1.
   Gravitational lens simulation
   Copyright (C) PG Sprimont 2002

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <stdio.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glut.h>

#include "gl.h"
#include "phys.h"
#include "../config.h"

//#define DEBUG

#define NANGLESMAX 600

#ifndef M_PI
#define M_PI 3.14159265
#endif

extern opt rt_opt;
extern view obs_view;


int ndefl=5;  //nombre max de dflecteurs
              //(changer pour en avoir plus)

float K=0.002; //Constante physique  (G/c^2 ou eqv)
float capture_radius=3.0; //rayon de capture des photons
int accurate=0; //accurate == on impose en plus que le photon
                //soit dans le plan

float *angles=NULL;
int nangles;

defl* dlist=NULL;

light* ray=NULL;
light* Ray=NULL;

int nrays=0;

src_param src_opt={0,0.75,1.0,{0.0,0.0}}; //voir .h

float step=0.2; //pas d'intgration

float id[16] = { 1,0,0,0,
        0,1,0,0,
        0,0,1,0,
        0,0,0,1};

/*-------------------------------------------------------------------------------*/
inline void norm(float* x)
/*-------------------------------------------------------------------------------*/
{
  register int i;
  float d=0;
  for(i=0;i<4;i++)
    d+=x[i]*x[i];

  d=sqrt(d);

  for(i=0;i<4;i++)
    x[i]/=d;

}
/*-------------------------------------------------------------------------------*/
inline void mult(GLfloat* M, GLfloat* v)
/*-------------------------------------------------------------------------------*/
{
  register int i, j;

  GLfloat n[4];

  for(i=0;i<4;i++)
	{
  	n[i]=0;
		for(j=0;j<4;j++)
			n[i]+=M[i*4+j]*v[j];
		for(j=0;j<4;j++)
			v[i]=n[i];
	}

}
/*-------------------------------------------------------------------------------*/
inline float vmult(GLfloat* v1, GLfloat* v2)
/*-------------------------------------------------------------------------------*/
{
  GLfloat d=0;
  register int i;
  for(i=0;i<4;i++)
    d+=v1[i]*v2[i];

  return d;
}
/*-------------------------------------------------------------------------------*/
void light::dev()   //calcule la dviation totale subie par un photon
{                   //sous l'effet de tous les dflecteurs
/*-------------------------------------------------------------------------------*/
		float a, b, c, cs, sn, rx, ry, rz;
		float dis, norm;
				
		int dontdo=0,k;

		GLfloat M[16];

		rx=ry=rz=0;

		for(k=0;k<ndefl;k++)
		{		
			a=b=c=0;

      if(dlist[k].active)
      {
			  a=(P[0]-dlist[k].P[0]);
			  b=(P[1]-dlist[k].P[1]);
			  c=(P[2]-dlist[k].P[2]);

			  dis = 1.0/sqrt(a*a+b*b+c*c);	
			
			  a*=K*dlist[k].m*1.0*dis; rx+=a;
			  b*=K*dlist[k].m*1.0*dis; ry+=b;
			  c*=K*dlist[k].m*1.0*dis; rz+=c;
      }
		}

	
		a=ry*DP[2]-rz*DP[1];
		b=rz*DP[0]-rx*DP[2];
		c=rx*DP[1]-ry*DP[0];

    norm=sqrt(a*a+b*b+c*c);
		dis=sqrt(rx*rx+ry*ry+rz*rz);
		
#ifdef DEBUG
//		printf("\nalpha=%f",norm);
#endif
		a/=norm;
		b/=norm;
		c/=norm;

		cs = cos(dis);
		sn = sin(dis);

//vrification dpassement systme
		if((isnan(a)) || (isnan(b)) || (isnan(c)))
		{
#ifdef DEBUG
				printf("!!NAN!!");
#endif
				dontdo=1;
		}
	
		M[0]=a*a*(1-cs)+cs;
		M[1]=a*b*(1-cs)-c*sn;
		M[2]=a*c*(1-cs)+b*sn;
		M[3]=0;
		M[4]=b*a*(1-cs)+c*sn;
		M[5]=b*b*(1-cs)+cs;
		M[6]=b*c*(1-cs)-a*sn;
		M[7]=0;
		M[8]=c*a*(1-cs)-b*sn;
		M[9]=c*b*(1-cs)+a*sn;
		M[10]=c*c*(1-cs)+cs;
		M[11]=0;
		M[12]=0;
		M[13]=0;
		M[14]=0;
		M[15]=1;

		if(!dontdo)
		  mult(M,DP);
		
		for(register int i=0;i<4;i++)
			P[i]+=DP[i]*step;

}
/*-------------------------------------------------------------------------------*/
void initdefl()   //Initialisation et dfinition des dflecteurs
/*-------------------------------------------------------------------------------*/
{
  int i,j;

  if(dlist!=NULL)	
   delete[] dlist;

	dlist = new defl[ndefl];

  for (i=0;i<ndefl;i++)
  {
    for(j=0;j<3;j++)
      dlist[i].P[j]=0.0;
    dlist[i].m=1.0;
    dlist[i].active=false;
    dlist[i].orbit=false;
  }

//dfinition de deux dflecteurs

	dlist[0].P[0]=rt_opt.Boxl/2;
	dlist[0].P[1]=rt_opt.Boxl/2;
	dlist[0].P[2]=-4*rt_opt.BoxL/5;
  dlist[0].m = 0.5;
  dlist[0].active=true;

	dlist[1].P[0]=rt_opt.Boxl/2;
	dlist[1].P[1]=rt_opt.Boxl/2;
	dlist[1].P[2]=-4*rt_opt.BoxL/5;
	dlist[1].m = 0.05;
  dlist[1].active=false;
//  dlist[1].orbit=true;

}
/*-------------------------------------------------------------------------------*/
void initray2() //Initialisation et dfinition de la source
                //Actuellement : source ponctuelle  l'infini
/*-------------------------------------------------------------------------------*/
{

  GLfloat srcX, srcY, srcZ;
  GLfloat x,y;
 	GLfloat dBoxl = src_opt.fraction*rt_opt.Boxl;

  if(ray!=NULL)
  	delete[] ray;
  if(Ray!=NULL)
  	delete[] Ray;

	int nx=(int)((dBoxl-1)*src_opt.density);

  #ifdef DEBUG	
	printf("\nNumber of rays : %d X %d = %d",nx,nx,nx*nx);
 	fflush(stdout);
  #endif


	int i,j;

	if(!src_opt.line)
	{
	nrays=(int) (nx*nx);
	ray = new light[nrays];
	Ray = new light[nrays];
	for(i=0;i<nx;i++)
		for(j=0;j<nx;j++)
		{
			x = rt_opt.Boxl/2-dBoxl/2+ (j+1)*dBoxl/(nx*1.0+1.0);
			y = rt_opt.Boxl/2-dBoxl/2+ (i+1)*dBoxl/(nx*1.0+1.0);

    ray[i*nx+j].P[0]= x;
		ray[i*nx+j].P[1]= y;
		ray[i*nx+j].P[2]=-rt_opt.BoxL;

		}
   }
	else
	{
		nrays= nx;
		ray = new light[nrays];
		Ray = new light[nrays];
		for(i=0;i<nx;i++)
		{
    	x = rt_opt.Boxl/2;
			y = rt_opt.Boxl/2-dBoxl/2+ (i+1)*dBoxl/(nx*1.0+1.0);

    ray[i].P[0]= x;
		ray[i].P[1]= y;
		ray[i].P[2]=-rt_opt.BoxL;
    	
    }

	}

  //direction du faisceau

  srcX=sin(src_opt.angle[0]/180.0*M_PI)*cos(src_opt.angle[1]/180.0*M_PI);
  srcY=sin(src_opt.angle[0]/180.0*M_PI)*sin(src_opt.angle[1]/180.0*M_PI);
  srcZ=cos(src_opt.angle[0]/180.0*M_PI);

	for(i=0;i<nrays;i++)
	{
		ray[i].DP[0]=srcX;
		ray[i].DP[1]=srcY;
		ray[i].DP[2]=srcZ;
    ray[i].flagged=false;
		Ray[i]=ray[i];
	}



  #ifdef DEBUG	
	printf("\nInitialisation rayons termine.");
 	fflush(stdout);
  #endif
}
/*-------------------------------------------------------------------------------*/
void restorerays() //Rinitialise les rayons dans leur position initiale
/*-------------------------------------------------------------------------------*/
{

	for(register int i=0;i<nrays;i++)
    for(register int j=0;j<3;j++)
		  {
        ray[i].P[j]=Ray[i].P[j];
        ray[i].DP[j]=Ray[i].DP[j];
      }

}
/*-------------------------------------------------------------------------------*/
void restoreraysall() //Rinitialise les rayons dans leur position initiale
/*-------------------------------------------------------------------------------*/
{
  nangles=0;
	for(register int i=0;i<nrays;i++)
    ray[i]=Ray[i];

}
/*-------------------------------------------------------------------------------*/
bool light::intersect() //Ce rayon intersecte-t-il ce plan ?
/*-------------------------------------------------------------------------------*/
{

  GLfloat d=0.0,dis=0.0;

  for(register int i=0;i<3;i++){
    d+=obs_view.a[i]*P[i];
    dis+=(P[i]-obs_view.z[i])*(P[i]-obs_view.z[i]);
  }

  d+=obs_view.a[3];

//une condition sur le plan, une sur la distance
// l'observateur

  if( (dis<=capture_radius*capture_radius) && (!accurate || (fabs(d) <= 2*step) ))
    {
#ifdef DEBUG
      printf("\nDis=%f\tEq=%f",sqrt(dis),d);
#endif
      getangle();
      return true;
     }
  return false;
}
/*-------------------------------------------------------------------------------*/
void view::update() //mets  jour l'quation du plan de l'observateur
/*-------------------------------------------------------------------------------*/
{
  GLfloat d=0.0;
  GLfloat ez[3]={0.0,0.0,1.0};

  mult(rot,ez);  //rotation de ez (norme ok, mat orth)

  for(register int i=0;i<3;i++)
    d+=(a[i]=ez[i])*z[i]; //calcul du terme inhomog.


  a[3]=-d;

}
/*-------------------------------------------------------------------------------*/
void view::identity() //rotation nulle
/*-------------------------------------------------------------------------------*/
{
  int i;
  for(i=0;i<16;i++)
    rot[i]=id[i];

}
/*-------------------------------------------------------------------------------*/
void light::getangle()
/*-------------------------------------------------------------------------------*/
{
  GLfloat e[3][4]={{1.0,0.0,0.0,0.0},
                   {0.0,1.0,0.0,0.0},
                   {0.0,0.0,1.0,0.0}};

  static bool stop=false;
  GLfloat x[4];

  for(register int i=2;i>-1;i--)
  {
    mult(obs_view.rot,e[i]); //changement de base
    x[i]=vmult(DP,e[i]); //projection
  }

  norm(x); //normalisation

  x[0]/=x[2]; //normalisation sur z
  x[1]/=x[2];

  if(!stop)
  {
    angles[2*nangles]=x[0];
    angles[2*nangles+1]=x[1];
    if(isnan(angles[2*nangles]) || isnan(angles[2*nangles+1]))
    {
      angles[2*nangles]=0.0;
      angles[2*nangles+1]=0.0;
    }
  }

  #ifdef DEBUG
  printf("\ntheta=%f\tPhi=%f",angles[2*nangles],angles[2*nangles+1]);
  #endif

  if(nangles==0)
    stop=false;

  if(nangles==NANGLESMAX/2-1)
    stop=true;
  else
    nangles++;
}
/*-------------------------------------------------------------------------------*/
int main (int argc, char** argv)   //main
/*-------------------------------------------------------------------------------*/
{
  angles=new float[NANGLESMAX];
  nangles=0;

  printf("\nHello, this is %s Version %s, an OpenGL", PACKAGE, VERSION);
  printf("\nGravitational lens simulator. Enjoy !\n");

	initdefl();
	initray2();

  glmain(argc, argv);

  if(dlist!=NULL)	
   delete[] dlist;
  if(ray!=NULL)
  	delete[] ray;
  if(Ray!=NULL)
  	delete[] Ray;
  if(angles!=NULL)
    delete[] angles;

  printf("\nGlens Finished !");
}
