/*
   glens
   Version 0.1.
   Gravitational lens simulation
   Copyright (C) PG Sprimont 2002

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <GL/glut.h>
#include <glui.h>

#include "ui.h"
#include "gl.h"
#include "../config.h"

#define PIX_SIZE  200
#define MOVING    210
#define VIEW_OBS  220
#define OBS_TRANSXY 230
#define OBS_TRANSZ 240
#define OBS_ROT 250
#define GLOB_TRANSX 251
#define GLOB_TRANSY 252
#define GLOB_TRANSZ 253
#define GLOB_ROT 254
#define B_CONFIG  260
#define B_3D 270
#define B_LOAD  280
#define B_SAVE  290
#define B_IDENT 300

#define BC_CLOSE 500

GLUI_Rotation *orot;
GLUI_Translation *otransxy, *otransz, *tx, *ty, *tz, *txy;
GLUI_Spinner *spin1;
GLUI_Checkbox *chk1;
GLUI_StaticText *t1, *t2;
GLUI_EditText* text;
GLUI_Button* but;
GLUI_Button     *b3d;
GLUI_Panel  *panelctr=0;

GLUI_RadioButton** radio=0;

GLUI *win_glui=0, *win_pos=0, *win_config=0, *win_control=0;


GLUI_StaticText **st;
GLUI_Checkbox   **cb;
GLUI_EditText   **ex;
GLUI_EditText   **ey;
GLUI_EditText   **ez;
GLUI_EditText   **em;


/*-------------------------------------------------------------------------------*/

extern int win_main;

extern opt rt_opt;
extern src_param src_opt;
extern float step;
extern int nseg;
extern float MassFactor;
extern float K;
extern float vuefrac;
extern float capture_radius;
extern int accurate;
extern float pixsize;

extern view glob_view;
extern view obs_view;
extern int ndefl;
extern defl* dlist;


/*-------------------------------------------------------------------------------*/

int moving=0;
float oldvuefrac=0;
int chk_showrays, chk_showbox;
bool config_open=false;

int d=0;
float x[3];

char fn[sizeof(GLUI_String)];

void save(char* filename)
{
  FILE* f;
  char* buf;
  long size;

  if(!(f=fopen(filename,"w+")))
    return;

  GLUI_Master.sync_live_all();

  buf = (char*) &rt_opt;
  size=sizeof(opt);
  fwrite(buf,size,1,f);

  buf = (char*) &src_opt;
  size=sizeof(src_param);
  fwrite(buf,size,1,f);

  buf = (char*) &step;
  size=sizeof(float);
  fwrite(buf,size,1,f);

  buf = (char*) &nseg;
  size=sizeof(int);
  fwrite(buf,size,1,f);

  buf = (char*) &MassFactor;
  size=sizeof(float);
  fwrite(buf,size,1,f);

  buf = (char*) &K;
  size=sizeof(float);
  fwrite(buf,size,1,f);

  buf = (char*) &MassFactor;
  size=sizeof(float);
  fwrite(buf,size,1,f);

  buf = (char*) &vuefrac;
  size=sizeof(float);
  fwrite(buf,size,1,f);

  buf = (char*) &capture_radius;
  size=sizeof(float);
  fwrite(buf,size,1,f);

  buf = (char*) &pixsize;
  size=sizeof(float);
  fwrite(buf,size,1,f);

  buf = (char*) &accurate;
  size=sizeof(int);
  fwrite(buf,size,1,f);

  buf = (char*) &glob_view;
  size=sizeof(view);
  fwrite(buf,size,1,f);

  buf = (char*) &obs_view;
  size=sizeof(view);
  fwrite(buf,size,1,f);

  buf = (char*) &ndefl;
  size=sizeof(int);
  fwrite(buf,size,1,f);

  buf = (char*) dlist;
  size=sizeof(defl);
  fwrite(buf,size,ndefl,f);


  fclose(f);


}

void load(char* filename)
{
  FILE* f;
  if(! (f=fopen(filename,"r")))
    return;

  char* buf;
  long size;

  buf = (char*) &rt_opt;
  size=sizeof(opt);
  fread(buf,size,1,f);

  buf = (char*) &src_opt;
  size=sizeof(src_param);
  fread(buf,size,1,f);

  buf = (char*) &step;
  size=sizeof(float);
  fread(buf,size,1,f);

  buf = (char*) &nseg;
  size=sizeof(int);
  fread(buf,size,1,f);

  buf = (char*) &MassFactor;
  size=sizeof(float);
  fread(buf,size,1,f);

  buf = (char*) &K;
  size=sizeof(float);
  fread(buf,size,1,f);

  buf = (char*) &MassFactor;
  size=sizeof(float);
  fread(buf,size,1,f);

  buf = (char*) &vuefrac;
  size=sizeof(float);
  fread(buf,size,1,f);

  buf = (char*) &capture_radius;
  size=sizeof(float);
  fread(buf,size,1,f);

  buf = (char*) &pixsize;
  size=sizeof(float);
  fread(buf,size,1,f);

  buf = (char*) &accurate;
  size=sizeof(int);
  fread(buf,size,1,f);

  buf = (char*) &glob_view;
  size=sizeof(view);
  fread(buf,size,1,f);

  buf = (char*) &obs_view;
  size=sizeof(view);
  fread(buf,size,1,f);

  buf = (char*) &ndefl;
  size=sizeof(int);
  fread(buf,size,1,f);

  buf = (char*) dlist;
  size=sizeof(defl);
  fread(buf,size,ndefl,f);

  fclose(f);

  GLUI_Master.sync_live_all();
  initray2();
  init(0,0);

}

void check()
{
  int i;
  if(radio)
  {
  for(i=0;i<ndefl;i++)
  {
  if(!dlist[i].active)
      radio[i]->disable();
    else
      radio[i]->enable();
  }
  }
}

void ctrl_callback(int ID)
{
  int i;

  switch(ID)
  {
    case 0:
      b3d->enable();
      win_control->close();
    break;
    case 1:
      for(i=0;i<3;i++) x[i]=dlist[d].P[i];
    break;
    case 2:
      for(i=0;i<3;i++) dlist[d].P[i]=x[i];
      win_config->sync_live();
    break;
    default:
    break;
  }

  if(ID >= 10)
  {
    for(i=0;i<3;i++) x[i]=dlist[d].P[i];

  }

}



/*-------------------------------------------------------------------------------*/
void control3d()
/*-------------------------------------------------------------------------------*/
{
  char text[256];
  int i;

  for(i=0;i<3;i++) x[i]=dlist[d].P[i];

  GLUI_Panel* panel;

  b3d->disable();
  win_control = GLUI_Master.create_glui("Controle 3D");

  panel = win_control->add_panel("Deflecteur:");
  panel->set_alignment(GLUI_ALIGN_LEFT);

  GLUI_RadioGroup *group = win_control->add_radiogroup_to_panel(panel, &d,1,ctrl_callback);

  radio = new (GLUI_RadioButton*)[ndefl];

  for(i=0;i<ndefl;i++)
  {
    sprintf(text,"%d",i);
    radio[i] = win_control->add_radiobutton_to_group(group,text);
  }

  check();

  win_control->add_column(false);

  panelctr = win_control->add_panel("Controle:");
  panelctr->set_alignment(GLUI_ALIGN_RIGHT);

  tx=win_control->add_translation_to_panel(panelctr,"X", GLUI_TRANSLATION_X, &dlist[d].P[0],2,ctrl_callback);
  win_control->add_column_to_panel(panel,false);
  ty=win_control->add_translation_to_panel(panelctr,"Y", GLUI_TRANSLATION_Y, &dlist[d].P[1],2,ctrl_callback);
  win_control->add_column_to_panel(panel,false);
  tz=win_control->add_translation_to_panel(panelctr,"Z", GLUI_TRANSLATION_Z, &dlist[d].P[2],2,ctrl_callback);
  win_control->add_column_to_panel(panel,false);
  txy=win_control->add_translation_to_panel(panelctr,"XY", GLUI_TRANSLATION_XY, &dlist[d].P[0],2,ctrl_callback);

  win_control->add_button("Close",0,ctrl_callback);

  delete radio;

}
/*-------------------------------------------------------------------------------*/
void cfg_callback(int ID)
/*-------------------------------------------------------------------------------*/
{
   int k;

   if(ID<2*ndefl-1)
   {
     if((ID%2)==0)
     {
        fflush(stdout);
        k = ID/2;

        if(dlist[k].active==true)
        {
          ex[k]->enable();
          ey[k]->enable();
          ez[k]->enable();
          em[k]->enable();
        }
          else
        {
          ex[k]->disable();
          ey[k]->disable();
          ez[k]->disable();
          em[k]->disable();
        }

        if(win_control)
        {
          win_control->sync_live();
          check();
        }

      }else
          control3d();
   }
   else
   if(ID==BC_CLOSE)
   {
      win_config->close();
      config_open=false;
      return;
   }

}

/*-------------------------------------------------------------------------------*/
void deflecteur_config()
/*-------------------------------------------------------------------------------*/
{
  config_open=true;
  win_config = GLUI_Master.create_glui("Configuration des dflecteurs");

  int n = ndefl; //Nombre de dflecteurs

  int i;

  st = new (GLUI_StaticText*)[n];
  cb = new (GLUI_Checkbox*)[n];
  ex = new (GLUI_EditText*)[n];
  ey = new (GLUI_EditText*)[n];
  ez = new (GLUI_EditText*)[n];
  em = new (GLUI_EditText*)[n];

  char text[10];

  for(i=0;i<n;i++)
  {
    sprintf(text,"Defl %d",i);
    st[i] = win_config->add_statictext(text);
    cb[i] = win_config->add_checkbox("Activer",(int*) &(dlist[i].active),2*i,cfg_callback);
    ex[i] = win_config->add_edittext("X", GLUI_EDITTEXT_FLOAT, &(dlist[i].P[0]));
    ey[i] = win_config->add_edittext("Y",GLUI_EDITTEXT_FLOAT, &(dlist[i].P[1]));
    ez[i] = win_config->add_edittext("Z",GLUI_EDITTEXT_FLOAT, &(dlist[i].P[2]));
    em[i] = win_config->add_edittext("Masse",GLUI_EDITTEXT_FLOAT, &(dlist[i].m));

    win_config->add_separator();
    win_config->add_column();

  }


  for(i=0;i<n;i++)
    if(dlist[i].active==false)
    {
       ex[i]->disable();
       ey[i]->disable();
       ez[i]->disable();
       em[i]->disable();
    }

  but=win_config->add_button("Close",BC_CLOSE,cfg_callback);

}



/*-------------------------------------------------------------------------------*/
void ui_callback(int ID)
/*-------------------------------------------------------------------------------*/
{

  switch(ID)
  {
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
      //initdefl();
      initray2();
      break;
    case PIX_SIZE:
      glPointSize(pixsize);
      break;
    case MOVING:
      if(moving)
      {         ;
       // otransxy->enable();
       // otransz->enable();
       // orot->enable();
        //chk1->enable();
       // spin1->enable();
      }
      else
      {    ;
        //otransxy->disable();
        //otransz->disable();
        //orot->disable();
        //chk1->disable();
        //spin1->disable();
      }
      break;
    case VIEW_OBS:
      if(rt_opt.dispobsview)
        vuefrac=oldvuefrac;
      else
      {
        oldvuefrac=vuefrac;
        vuefrac=1.0;
      }
      break;
    case OBS_TRANSXY:
    case OBS_TRANSZ:
    case OBS_ROT:
      restoreraysall();
      obs_view.update();
       break;
    case GLOB_TRANSX:
    case GLOB_TRANSY:
    case GLOB_TRANSZ:
    case GLOB_ROT:
       //win_pos->sync_live();
       break;
    case B_CONFIG:
      if(!config_open)
        deflecteur_config();
      break;
    case B_3D:
      control3d();
      break;
    case B_SAVE:
      save(fn);
      break;
    case B_LOAD:
      load(fn);
      break;
    case B_IDENT:
      glob_view.identity();
      win_glui->sync_live();
      break;
    default:
      break;
  }
}




/*-------------------------------------------------------------------------------*/
void init_glui()
/*-------------------------------------------------------------------------------*/
{

  sprintf(fn,"glens.cgf");

  win_glui = GLUI_Master.create_glui_subwindow(win_main, GLUI_SUBWINDOW_RIGHT);
  win_glui->set_main_gfx_window(win_main);

  t1=win_glui->add_statictext("Glens!");
  t2=win_glui->add_statictext("mailto: fullmoon@swing.be");

  t1->set_alignment(GLUI_ALIGN_CENTER);
  t2->set_alignment(GLUI_ALIGN_CENTER);

  win_glui->add_separator();

/*-------------------------------------------------------------------------------*/
  GLUI_Rollout* panel3 = win_glui->add_rollout("Source",0);
/*-------------------------------------------------------------------------------*/

  panel3->set_alignment(GLUI_ALIGN_LEFT);

  GLUI_RadioGroup *group = win_glui->add_radiogroup_to_panel(panel3, &src_opt.line,3,ui_callback);
  win_glui->add_radiobutton_to_group(group,"Matrice");
  win_glui->add_radiobutton_to_group(group,"Ligne");

  win_glui->add_separator_to_panel(panel3);

  text =win_glui->add_edittext_to_panel(panel3, "Densite", GLUI_EDITTEXT_FLOAT, &(src_opt.density),6,ui_callback);
  text->set_float_limits(0.0,50.0,GLUI_LIMIT_CLAMP);
  text=win_glui->add_edittext_to_panel(panel3, "Fraction", GLUI_EDITTEXT_FLOAT, &(src_opt.fraction),7,ui_callback);
  text->set_float_limits(0.0,10.0,GLUI_LIMIT_CLAMP);

/*-------------------------------------------------------------------------------*/
  GLUI_Rollout* panel4 = win_glui->add_rollout_to_panel(panel3,"Orientation",0);
/*-------------------------------------------------------------------------------*/

  panel4->set_alignment(GLUI_ALIGN_LEFT);

  GLUI_Spinner* spin =win_glui->add_spinner_to_panel(panel4, "Theta", GLUI_EDITTEXT_FLOAT, &(src_opt.angle[0]),4,ui_callback);
  spin->set_float_limits(-90.0,90.0,GLUI_LIMIT_CLAMP);
  spin=win_glui->add_spinner_to_panel(panel4, "Phi", GLUI_EDITTEXT_FLOAT, &(src_opt.angle[1]),5,ui_callback);
  spin->set_float_limits(0.0,360.0,GLUI_LIMIT_WRAP);

/*-------------------------------------------------------------------------------*/
  GLUI_Rollout* panel6 = win_glui->add_rollout("Deflecteur",0);
/*-------------------------------------------------------------------------------*/

  panel6->set_alignment(GLUI_ALIGN_LEFT);

  win_glui->add_button_to_panel(panel6, "Configuration",B_CONFIG,ui_callback);
//  b3d = win_glui->add_button_to_panel(panel6,"Controle 3D",B_3D,ui_callback);

/*-------------------------------------------------------------------------------*/
  GLUI_Rollout* panel2 = win_glui->add_rollout("Zone de calcul",0);
/*-------------------------------------------------------------------------------*/

  panel2->set_alignment(GLUI_ALIGN_LEFT);

  text =win_glui->add_edittext_to_panel(panel2, "Largeur", GLUI_EDITTEXT_FLOAT, &(rt_opt.Boxl),1,ui_callback);
  text->set_float_limits(0.0,1000.0,GLUI_LIMIT_CLAMP);
  text=win_glui->add_edittext_to_panel(panel2, "Longueur", GLUI_EDITTEXT_FLOAT, &(rt_opt.BoxL),2,ui_callback);
  text->set_float_limits(0.0,1000.0,GLUI_LIMIT_CLAMP);
  win_glui->add_separator_to_panel(panel2);
  win_glui->add_checkbox_to_panel(panel2, "Limite ZC lateral", &(rt_opt.boundary_lateral));
  win_glui->add_checkbox_to_panel(panel2, "Limite ZC front", &(rt_opt.boundary_front));

/*-------------------------------------------------------------------------------*/
  GLUI_Rollout* panel5 = win_glui->add_rollout_to_panel(panel2,"Integration",0);
/*-------------------------------------------------------------------------------*/

  panel5->set_alignment(GLUI_ALIGN_LEFT);

  spin =win_glui->add_spinner_to_panel(panel5, "Intensite", GLUI_EDITTEXT_FLOAT, &(K),10,ui_callback);
  spin->set_float_limits(0.0,100.0,GLUI_LIMIT_CLAMP);
  spin->set_speed(0.0001);
  text =win_glui->add_edittext_to_panel(panel5, "Pas", GLUI_EDITTEXT_FLOAT, &(step),8,ui_callback);
  text->set_float_limits(0.0,10.0,GLUI_LIMIT_CLAMP);
  text=win_glui->add_edittext_to_panel(panel5, "NMax/ray", GLUI_EDITTEXT_INT, &(nseg),9,ui_callback);
  text->set_int_limits(0,10000,GLUI_LIMIT_CLAMP);

/*-------------------------------------------------------------------------------*/
  GLUI_Rollout* panel7 = win_glui->add_rollout("Observateur",0);
/*-------------------------------------------------------------------------------*/

  panel7->set_alignment(GLUI_ALIGN_LEFT);

  win_glui->add_checkbox_to_panel(panel7, "Activation", &(moving), MOVING, ui_callback);
  spin1=spin =win_glui->add_spinner_to_panel(panel7, "Rayon de capture", GLUI_EDITTEXT_FLOAT, &(capture_radius));
  spin->set_float_limits(0.0,50.0,GLUI_LIMIT_CLAMP);
  spin->set_speed(1);
  win_glui->add_separator_to_panel(panel7);
  chk1=win_glui->add_checkbox_to_panel(panel7, "Plan", &(accurate));

  //chk1->disable();
  //spin1->disable();

/*-------------------------------------------------------------------------------*/
  GLUI_Rollout* panel1 = win_glui->add_rollout("Affichage",0);
/*-------------------------------------------------------------------------------*/

  panel1->set_alignment(GLUI_ALIGN_LEFT);

  win_glui->add_checkbox_to_panel(panel1, "Rayons", &(rt_opt.disprays));
  win_glui->add_checkbox_to_panel(panel1,"Zone de calcul", &(rt_opt.dispbox));
  win_glui->add_checkbox_to_panel(panel1,"Axes", &(rt_opt.dispaxes));
  win_glui->add_checkbox_to_panel(panel1,"Deflecteurs", &(rt_opt.dispdefl));
  win_glui->add_checkbox_to_panel(panel1,"Vue Observateur", &(rt_opt.dispobsview),VIEW_OBS,ui_callback);
  spin =win_glui->add_spinner_to_panel(panel1, "Split Obs/Main", GLUI_EDITTEXT_FLOAT, &(vuefrac));
  spin->set_float_limits(0.0,1.0,GLUI_LIMIT_CLAMP);
  spin->set_speed(1);
  spin =win_glui->add_spinner_to_panel(panel1, "Taille PixelObs", GLUI_EDITTEXT_FLOAT, &(pixsize), PIX_SIZE, ui_callback);
  spin->set_float_limits(0.0,10.0,GLUI_LIMIT_CLAMP);
  spin->set_speed(1);
  win_glui->add_checkbox_to_panel(panel1, "Stats graphiques", &(rt_opt.dispstats));


/*-------------------------------------------------------------------------------*/
  win_pos = GLUI_Master.create_glui_subwindow(win_main, GLUI_SUBWINDOW_BOTTOM);
/*-------------------------------------------------------------------------------*/

//  win_glui->set_main_gfx_window(win_main);
  GLUI_Panel* p = win_pos->add_panel("Global",0);

  win_pos->add_rotation_to_panel(p,"Global",glob_view.rot,GLOB_ROT,ui_callback);

  win_pos->add_column_to_panel(p,false);
  win_pos->add_translation_to_panel(p,"Global X", GLUI_TRANSLATION_X, &glob_view.z[0],GLOB_TRANSX,ui_callback);


//  spin =win_pos->add_spinner("", GLUI_EDITTEXT_FLOAT, &(glob_view.z[0]));
//  spin->set_float_limits(-1000.0,1000.0,GLUI_LIMIT_CLAMP);
//  spin->set_speed(0.01);

  win_pos->add_column_to_panel(p,false);
  win_pos->add_translation_to_panel(p,"Global Y", GLUI_TRANSLATION_Y, &glob_view.z[1],GLOB_TRANSY,ui_callback);
//  spin =win_pos->add_spinner("", GLUI_EDITTEXT_FLOAT, &(glob_view.z[1]));
//  spin->set_float_limits(-1000.0,1000.0,GLUI_LIMIT_CLAMP);
//  spin->set_speed(0.01);

  win_pos->add_column_to_panel(p,false);
  win_pos->add_translation_to_panel(p,"Global Z", GLUI_TRANSLATION_Z, &glob_view.z[2],GLOB_TRANSZ,ui_callback);
//  spin =win_pos->add_spinner("", GLUI_EDITTEXT_FLOAT, &(glob_view.z[2]));
//  spin->set_float_limits(-1000.0,1000.0,GLUI_LIMIT_CLAMP);
//  spin->set_speed(0.01);

  win_pos->add_column(true);

  p = win_pos->add_panel("Observer",0);
  orot = win_pos->add_rotation_to_panel(p,"Obs Rot",obs_view.rot,OBS_ROT, ui_callback);
  win_pos->add_column_to_panel(p,false);
  otransxy = win_pos->add_translation_to_panel(p,"Obs X", GLUI_TRANSLATION_X, &obs_view.z[0],OBS_TRANSXY,ui_callback);
  win_pos->add_column_to_panel(p,false);
  otransxy = win_pos->add_translation_to_panel(p,"Obs Y", GLUI_TRANSLATION_Y, &obs_view.z[1],OBS_TRANSXY,ui_callback);
  win_pos->add_column_to_panel(p,false);
  otransz = win_pos->add_translation_to_panel(p,"Obs Z", GLUI_TRANSLATION_Z, &obs_view.z[2],OBS_TRANSZ,ui_callback);

  //otransxy->disable();
  //otransz->disable();
  //orot->disable();
  win_pos->add_column(false);
  win_pos->add_button("Identite",B_IDENT,ui_callback);


  win_glui->add_edittext("Fichier:", GLUI_EDITTEXT_TEXT, fn);

  but=win_glui->add_button( "Sauver config", B_SAVE,ui_callback );
  but->set_alignment(GLUI_ALIGN_LEFT);
  but=win_glui->add_button( "Charger config", B_LOAD,ui_callback );
  but->set_alignment(GLUI_ALIGN_LEFT);
  but=win_glui->add_button( "Quitter", 0,(GLUI_Update_CB) exit );
  but->set_alignment(GLUI_ALIGN_LEFT);

  load(fn);

}

